/**
 * Copyright (c) 2021, WSO2 Inc. (http://www.wso2.com). All Rights Reserved.
 *
 * This software is the property of WSO2 Inc. and its suppliers, if any.
 * Dissemination of any information or reproduction of any material contained
 * herein in any form is strictly forbidden, unless permitted by WSO2 expressly.
 * You may not alter or remove any copyright or other notice from copies of this content."
 */

import { IdentifiableComponentInterface } from "@wso2is/core/models";
import { Heading, Link, ResourceTab, Text } from "@wso2is/react-components";
import cloneDeep from "lodash-es/cloneDeep";
import isEmpty from "lodash-es/isEmpty";
import set from "lodash-es/set";
import React, { FunctionComponent, ReactElement, useEffect, useState } from "react";
import { Trans, useTranslation } from "react-i18next";
import { useSelector } from "react-redux";
import { Divider, Icon, Message, Segment, TabProps } from "semantic-ui-react";
import {
    AdvanceForm,
    AdvanceFormValuesInterface,
    DesignForm,
    DesignFormValuesInterface,
    GeneralDetailsForm,
    GeneralDetailsFormValuesInterface
} from "./forms";
import { BrandingPreferencePreview } from "./preview";
import { AppState } from "../../../../features/core/store";
import { BrandingPreferencesConstants } from "../constants";
import { BrandingPreferenceMeta } from "../meta";
import { BrandingPreferenceInterface } from "../models";

/**
 * Proptypes for the Branding preference tabs component.
 */
interface BrandingPreferenceTabsInterface extends IdentifiableComponentInterface {
    /**
     * Branding preferences object.
     */
    brandingPreference: BrandingPreferenceInterface;
    /**
     * Is the component loading.
     */
    isLoading: boolean;
    /**
     * Is Read only?
     */
    readOnly: boolean;
    /**
     * Should the tabs be split in to two views.
     */
    isSplitView: boolean;
    /**
     * Is the component updating.
     */
    isUpdating: boolean;
    /**
     *
     * @param {GeneralDetailsFormValuesInterface
     *  | DesignFormValuesInterface
     *  | AdvanceFormValuesInterface } values - Form Values.
     */
    onSubmit: (values: GeneralDetailsFormValuesInterface
        | DesignFormValuesInterface
        | AdvanceFormValuesInterface) => void;
}

/**
 * Branding Preference Tabs.
 *
 * @param {BrandingPreferenceTabsInterface} props - Props injected to the component.
 *
 * @return {React.ReactElement}
 */
export const BrandingPreferenceTabs: FunctionComponent<BrandingPreferenceTabsInterface> = (
    props: BrandingPreferenceTabsInterface
): ReactElement => {

    const {
        ["data-componentid"]: componentId,
        brandingPreference,
        isLoading,
        isSplitView,
        isUpdating,
        readOnly,
        onSubmit
    } = props;

    const { t } = useTranslation();

    const supportEmail: string = useSelector((state: AppState) =>
        state.config.deployment.extensions?.supportEmail as string);
    const theme: string = useSelector((state: AppState) => state.config.ui.theme?.name);

    const [ isSubmitting, setIsSubmitting ] = useState<boolean>(isUpdating);
    const [
        themeColors,
        setThemeColors
    ] = useState<BrandingPreferenceInterface[ "colors" ]>(BrandingPreferencesConstants.DEFAULT_PREFERENCE.colors);
    const [
        brandingPreferenceForPreview,
        setBrandingPreferenceForPreview
    ] = useState<BrandingPreferenceInterface>(brandingPreference);

    /**
     * Resolves the theme colors on component mount.
     */
    useEffect(() => {

        resolveThemeColors();
    }, []);

    /**
     * Sets the branding preference preview config.
     */
    useEffect(() => {

        if (isEmpty(brandingPreference)) {
            return;
        }

        setBrandingPreferenceForPreview(moderateValuesForPreview(brandingPreference));
    }, [ brandingPreference ]);

    /**
     * Sets the submitting state of the form.
     */
    useEffect(() => {

        setIsSubmitting(isUpdating);
    }, [ isUpdating ]);

    /**
     * Fetches the theme variables file and resolves the color palette of the theme.
     */
    const resolveThemeColors = (): void => {

        import(`../../../../themes/${ theme }/theme-variables.json`)
            .then((response: Record<string, string>) => {
                setThemeColors({
                    primary: response.primaryColor
                });
            })
            .catch(() => {
                // Add debug logs here one a logger is added.
                // Tracked here https://github.com/wso2/product-is/issues/11650.
            });
    };

    /**
     * General Preferences Tab.
     * @return {ReactElement}
     */
    const GeneralPreferenceTabPane = (): ReactElement => (
        <ResourceTab.Pane loading={ isLoading } attached="bottom">
            <Segment basic padded="very" className="preference-form-container">
                <GeneralDetailsForm
                    onSubmit={ onSubmit }
                    initialValues={ {
                        organizationDetails: {
                            copyrightText: brandingPreference.organizationDetails?.copyrightText,
                            siteTitle: brandingPreference.organizationDetails?.siteTitle,
                            supportEmail: brandingPreference.organizationDetails?.supportEmail
                        }
                    } }
                    isSubmitting={ isSubmitting }
                    broadcastValues={ (values: GeneralDetailsFormValuesInterface) => {
                        setBrandingPreferenceForPreview({
                            ...moderateValuesForPreview(brandingPreference),
                            organizationDetails: {
                                copyrightText: !isEmpty(values.organizationDetails.copyrightText)
                                    ? values.organizationDetails.copyrightText
                                    : BrandingPreferenceMeta
                                        .getBrandingPreferenceFallbacks().organizationDetails.copyrightText,
                                siteTitle: values.organizationDetails.siteTitle,
                                supportEmail: values.organizationDetails.supportEmail
                            }
                        });
                    } }
                    readOnly={ readOnly }
                />
            </Segment>
        </ResourceTab.Pane>
    );

    /**
     * Design Preferences Tab.
     * @return {ReactElement}
     */
    const DesignPreferenceTabPane = (): ReactElement => (
        <ResourceTab.Pane loading={ isLoading } attached="bottom">
            <Segment basic padded="very" className="preference-form-container">
                <DesignForm
                    onSubmit={ onSubmit }
                    initialValues={ {
                        colors: {
                            primary: brandingPreference.colors?.primary ?? themeColors.primary
                        },
                        images: {
                            favicon: {
                                imgURL: brandingPreference.images?.favicon?.imgURL
                            },
                            logo: {
                                altText: brandingPreference.images?.logo?.altText,
                                imgURL: brandingPreference.images?.logo?.imgURL
                            }
                        }
                    } }
                    isSubmitting={ isSubmitting }
                    broadcastValues={ (values: DesignFormValuesInterface) => {
                        setBrandingPreferenceForPreview({
                            ...moderateValuesForPreview(brandingPreference),
                            colors: values.colors,
                            images: {
                                favicon: {
                                    ...values.images.favicon,
                                    imgURL: !isEmpty(values.images.favicon.imgURL)
                                        ? values.images.favicon.imgURL
                                        : BrandingPreferenceMeta.getBrandingPreferenceFallbacks().images.favicon.imgURL
                                },
                                logo: {
                                    ...values.images.logo,
                                    imgURL: !isEmpty(values.images.logo.imgURL)
                                        ? values.images.logo.imgURL
                                        : BrandingPreferenceMeta.getBrandingPreferenceFallbacks().images.logo.imgURL
                                }
                            }
                        });
                    } }
                    readOnly={ readOnly }
                />
            </Segment>
        </ResourceTab.Pane>
    );

    /**
     * Advance Preferences Tab.
     * @return {ReactElement}
     */
    const AdvancePreferenceTabPane = (): ReactElement => (
        <ResourceTab.Pane loading={ isLoading } attached="bottom">
            <Segment basic padded="very" className="preference-form-container">
                <AdvanceForm
                    onSubmit={ onSubmit }
                    initialValues={ {
                        urls: {
                            cookiePolicyURL: brandingPreference.urls?.cookiePolicyURL,
                            privacyPolicyURL: brandingPreference.urls?.privacyPolicyURL,
                            termsOfUseURL: brandingPreference.urls?.termsOfUseURL
                        }
                    } }
                    isSubmitting={ isSubmitting }
                    broadcastValues={ (values: AdvanceFormValuesInterface) => {
                        setBrandingPreferenceForPreview({
                            ...moderateValuesForPreview(brandingPreference),
                            urls: values.urls
                        });
                    } }
                    readOnly={ readOnly }
                />
                <Divider hidden/>
                <Heading as="h3">{ t("extensions:develop.branding.tabs.general.customRequest.heading") }</Heading>
                <Text>
                    <Trans
                        i18nKey={ "extensions:develop.branding.tabs.general.customRequest.description" }
                        tOptions={ {
                            supportEmail
                        } }
                    >
                        If you require further customizations, please reach to us at <Link
                            link={ `mailto:${ supportEmail }` }>{ supportEmail }</Link>
                    </Trans>
                </Text>
            </Segment>
        </ResourceTab.Pane>
    );

    /**
     * Preview Preferences Tab.
     * @return {ReactElement}
     */
    const PreviewPreferenceTabPane = (): ReactElement => (
        <ResourceTab.Pane className="preview-tab" attached="bottom">
            <Message
                info
                floating
                attached="top"
                className="preview-disclaimer"
                content={ (
                    <>
                        <Icon name="info circle"/>
                        { t("extensions:develop.branding.tabs.preview.disclaimer") }
                    </>
                ) }
            />
            <BrandingPreferencePreview
                isLoading={ isLoading }
                brandingPreference={ brandingPreferenceForPreview }
            />
        </ResourceTab.Pane>
    );

    /**
     * Resolves the tab panes based on the application config.
     *
     * @return {TabProps [ "panes" ]} Resolved tab panes.
     */
    const resolveTabPanes = (): TabProps [ "panes" ] => {

        const panes: TabProps [ "panes" ] = [];

        panes.push({
            menuItem: t("extensions:develop.branding.tabs.general.label"),
            render: GeneralPreferenceTabPane
        });

        panes.push({
            menuItem: t("extensions:develop.branding.tabs.design.label"),
            render: DesignPreferenceTabPane
        });

        panes.push({
            menuItem: t("extensions:develop.branding.tabs.advance.label"),
            render: AdvancePreferenceTabPane
        });
        
        if (!isSplitView) {
            panes.push({
                menuItem: t("extensions:develop.branding.tabs.preview.label"),
                render: PreviewPreferenceTabPane
            });
        }

        return panes;
    };

    /**
     * Moderate the values for the preview window.
     * @param {BrandingPreferenceInterface} preference - Preference Object.
     */
    const moderateValuesForPreview = (preference: BrandingPreferenceInterface): BrandingPreferenceInterface => {

        const preferenceForPreview: BrandingPreferenceInterface = cloneDeep(preference);

        if (isEmpty(brandingPreference.images?.logo?.imgURL)) {
            set(preferenceForPreview,
                "images.logo.imgURL",
                BrandingPreferenceMeta.getBrandingPreferenceFallbacks().images.logo.imgURL);
        }

        if (isEmpty(brandingPreference.images?.favicon?.imgURL)) {
            set(preferenceForPreview,
                "images.favicon.imgURL",
                BrandingPreferenceMeta.getBrandingPreferenceFallbacks().images.favicon.imgURL);
        }

        if (isEmpty(brandingPreference.organizationDetails?.copyrightText)) {
            set(preferenceForPreview,
                "organizationDetails.copyrightText",
                BrandingPreferenceMeta.getBrandingPreferenceFallbacks().organizationDetails.copyrightText);
        }

        if (isEmpty(brandingPreference.colors?.primary)) {
            set(preferenceForPreview, "colors.primary", themeColors?.primary);
        }

        return preferenceForPreview;
    };

    return (
        <Segment.Group
            horizontal
            className="basic branding-preference-tab-group"
            data-componentid={ componentId }
        >
            <Segment basic padded={ false }>
                <ResourceTab
                    attached="top"
                    secondary={ false }
                    pointing={ false }
                    onTabChange={ () => setBrandingPreferenceForPreview(moderateValuesForPreview(brandingPreference)) }
                    panes={ resolveTabPanes() }
                />
            </Segment>
            {
                isSplitView && (
                    <Segment basic padded={ false } className="preview">
                        <ResourceTab
                            attached="top"
                            secondary={ false }
                            pointing={ false }
                            panes={ [
                                {
                                    menuItem: t("extensions:develop.branding.tabs.preview.label"),
                                    render: PreviewPreferenceTabPane
                                }
                            ] }
                        />
                    </Segment>
                )
            }
        </Segment.Group>
    );
};

/**
 * Default props for the component.
 */
BrandingPreferenceTabs.defaultProps = {
    "data-componentid": "branding-preference-tabs"
};
